/****
   Modification History:
   [1.0]    07.09.2009    Tobias Pluess
   - first release

   [1.1]    03.11.2009    Tobias Pluess
   - added initializations for the vic and the gpio
   - added initialization for the memory mapping and the mam
 *****************************************************************************/

/******************************************************************************
 * INCLUDE FILES
 *****************************************************************************/

#include "cstart.h"

/******************************************************************************
 * PRIVATE FUNCTION PROTOTYPES (STATIC)
 *****************************************************************************/

/******************************************************************************
 * PRIVATE VARIABLES
 *****************************************************************************/

/******************************************************************************
 * OTHER DEFINITIONS
 *****************************************************************************/

#define OSC_RANGE_SEL OSCRANGE_LO /* low oscillator range */
#define CLOCKSOURCE CLK_MAIN /* select the clock source for the pll */
#define CCLK_DIV 3 /* set the cpu clock divider - see NXP PLL Excel sheet */
#define NSEL 1 /* N value for pll - see NXP PLL Excel sheet */
#define MSEL 36 /* M value for pll - see NXP PLL Excel sheet */

/******************************************************************************
 * MODULE FUNCTIONS (PUBLIC)
 *****************************************************************************/

/*===========================================================================*/
tDWord fDispatchSWI(tDWord dwSWINbr)
/*-----------------------------------------------------------------------------
  Function:
  dispatcher for software interrupts (swi #)
  in:  dwSWINbr -> parameter of the swi call (24 bits)
  out: return code
=============================================================================*/
{
  return 0;
} /* end fDispatchSWI */

/*===========================================================================*/
void fLowLevelInit(void)
/*-----------------------------------------------------------------------------
  Function:
  low-level hardware initialization function; will be called directly after
  reset.
  in:  none
  out: none
=============================================================================*/
{
  volatile tDWord dwDelay;

#ifdef DEBUG
  MEMMAP = 2; /* map interrupt vectors to internal RAM (debug in RAM) */
#else
  MEMMAP = 1; /* no remapping of interrupt vectors */
#endif

  /* initialize the pll */
#if CLOCKSOURCE == CLK_MAIN /* check if the main osc is the clk source */
  SCS |= OSCRANGE; /* first select an oscillator range */
  SCS |= OSCEN; /* then enable the main oscillator */
  while(!(SCS & OSCSTAT)); /* wait until the main osc is ready */
#endif
  PLLCON &= ~PLLC; /* disconnect the pll */
  PLLFEED = 0xAA;
  PLLFEED = 0x55;
  PLLCON &= ~PLLE; /* disable the pll */
  PLLFEED = 0xAA;
  PLLFEED = 0x55;
  CCLKCFG = 0; /* disable cpu clock divider to speed up pll configuration */
  CLKSRCSEL = CLOCKSOURCE; /* select the desired clock source */

  /* set the N and M values */
  PLLCFG = (((NSEL - 1) & 0xFF) << 16) | ((MSEL - 1) & 0xFFFF);
  PLLFEED = 0xAA;
  PLLFEED = 0x55;
  PLLCON |= PLLE; /* enable the pll */
  PLLFEED = 0xAA;
  PLLFEED = 0x55;
  CCLKCFG = CCLK_DIV; /* select cpu clock divider */
  while(!(PLLSTAT & PLOCK)); /* wait for the pll to lock */
  PLLCON |= PLLC; /* connect the pll */
  PLLFEED = 0xAA;
  PLLFEED = 0x55;

  VICIntEnClr = 0xFFFFFFFF; /* initialize the vic by clearing any interrupts */
  VICAddress = 0; /* reset vic address register */
  VICIntSelect = 0; /* switch all interrupts to irq mode */

  MAMTIM = 3; /* initialize the memory accelerator module */
  MAMCR = 2;

#ifdef HI_SPEED_GPIO
  SCS |= BIT_00; /* select all gpios as fio instead of legacy */
#else
  SCS &= ~BIT_00; /* select legacy io ports */
#endif

  PCONP |= BIT_11; /* enable power for the emc */
  PINSEL5 = 0x55010115; /* enable DQMx, CKEOUT0, DYCS0, CLKOUT0, RAS, CAS */
  PINMODE5 = 0xAA02022A; /* disable the pullups/pulldowns */
  PINSEL6 = 0x55555555; /* enable data bus pins D15..D0 */
  PINMODE6 = 0xFFFFFFFF; /* activate the pulldowns for the data bus */
  PINSEL7 = 0x55555555; /* enable data bus pins D31..D16 */
  PINMODE7 = 0xFFFFFFFF; /* activate the pulldowns for the data bus */
  PINSEL8 = 0x55555555; /* address bus pins A15..A0 */
  PINMODE8 = 0xAAAAAAAA; /* disable the pullups/pulldowns for address bus */
  PINSEL9 = 0x50150555; //0x10150555; /* enable the CS0, BLS0, WE, OE, A21..A16 */
  PINMODE9 = 0xA02A0AAA; //0x202A0AAA; /* disable the pulldowns for all control signals */
  EMCControl = BIT_00; /* disable the emc before the registers are set up */

  /* set up the flash memory */
  EMCStaticConfig0 = BIT_00 | BIT_07;
  EMCStaticWaitWen0 = 0;
  EMCStaticWaitOen0 = 0;
  EMCStaticWaitRd0 = 7;
  EMCStaticWaitPage0 = 0;
  EMCStaticWaitWr0 = 7;
  EMCStaticWaitTurn0 = 0;

  /* set up the fpga memory bus */
  EMCStaticConfig1 = 0;
  EMCStaticWaitWen1 = 0;
  EMCStaticWaitOen1 = 0;
  EMCStaticWaitRd1 = 2;
  EMCStaticWaitPage1 = 0;
  EMCStaticWaitWr1 = 2;
  EMCStaticWaitTurn1 = 0;

  /* set up the sdram */
  EMCDynamicReadConfig = 1; /* select command delayed read strategy */
  EMCDynamicRasCas0 = 0x203; /* cas latency 2, ras latency 3 */
  EMCDynamictRP = 1; /* see sdram timing calculation sheet */
  EMCDynamictRAS = 3;
  EMCDynamictSREX = 5;
  EMCDynamictAPR = 1;
  EMCDynamictDAL = 2;
  EMCDynamictWR = 4;
  EMCDynamictRC = 4;
  EMCDynamictRFC = 4;
  EMCDynamictXSR = 5;
  EMCDynamictRRD = 1;
  EMCDynamictMRD = 2;

  /* configure the emc for
  256 MB (16Mx16), 4 banks, row length = 13, column length = 9 */
  EMCDynamicConfig0 = (BIT_14 | BIT_10 | BIT_09 | BIT_07);
  for(dwDelay = 0; dwDelay < 100; dwDelay++);
  EMCDynamicControl = 0x181; /* issue nop command */
  for(dwDelay = 0; dwDelay < 100; dwDelay++);
  EMCDynamicControl = 0x101; /* issue precharge all command */
  EMCDynamicRefresh = 1; /* start the refresh timer */
  for(dwDelay = 0; dwDelay < 100; dwDelay++);
  EMCDynamicRefresh = 30; /*19 set up the refresh timer */
  EMCDynamicControl = 0x081; /* issue sdram mode command */
  dwDelay = (*(tpDWord)(0xA0000000 | (0x22 << 13))); /* burst length 8, cl 2 */
  EMCDynamicControl = BIT_01; /* clkout runs continuously */
  EMCDynamicConfig0 |= BIT_19; /* enable the buffer */
} /* end fLowLevelInit */

/******************************************************************************
 * PRIVATE FUNCTIONS (STATIC)
 *****************************************************************************/

/******************************************************************************
 * END OF CODE
 *****************************************************************************/
