#include "base_def.h"
#include "lwip/opt.h"

#include "lwip/def.h"
#include "lwip/mem.h"
#include "lwip/pbuf.h"
#include "lwip/sys.h"
#include "lwip/stats.h"
#include "lwip/snmp.h"
#include "netif/etharp.h"
#include "netif/ppp_oe.h"
#include "netif/ethernetif.h"

static err_t
ethernetif_low_level_output(struct netif *netif, struct pbuf *p);

static void
ethernetif_linkinterrupt(void);

static void
ethernetif_interrupt(void);

static void
init_tx_descriptors(void);

static void
init_rx_descriptors(void);

static void
ethernetif_write_phy(tWord wReg, tWord wValue);

static u16_t
ethernetif_read_phy(u16_t wReg);

static struct pbuf*
low_level_input(void);

#pragma data_alignment = 4
static u8_t rx_buffer[MAC_RX_FRAGMENTS * ETH_FRAG_SIZE];

#pragma location = "EMAC_DMA_RAM"
static u32_t rx_descriptors[MAC_RX_FRAGMENTS * 2];

#pragma location = "EMAC_DMA_RAM"
static u32_t rx_status_descriptors[MAC_RX_FRAGMENTS * 2];

#pragma data_alignment = 4
static u8_t tx_buffer[MAC_TX_FRAGMENTS * ETH_FRAG_SIZE];

#pragma location = "EMAC_DMA_RAM"
static u32_t tx_descriptors[MAC_TX_FRAGMENTS * 2];

#pragma location = "EMAC_DMA_RAM"
static u32_t tx_status_descriptors[MAC_TX_FRAGMENTS];

volatile enum netif_status* status;

err_t
ethernetif_init(struct netif *netif)
{
  netif->hwaddr_len = ETHARP_HWADDR_LEN; /* set MAC hardware address length */
  SET_MAC(netif->hwaddr); /* set MAC hardware address */
  netif->mtu = ETH_FRAG_SIZE; /* maximum transfer unit */
  netif->flags = NETIF_FLAG_LINK_UP | NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP; /* device capabilities */
  netif->output = etharp_output;
  netif->linkoutput = ethernetif_low_level_output;
#ifdef LWIP_NETIF_HOSTNAME
  netif->hostname = LWIP_HOSTNAME;
#endif
  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;
  status = netif->state;

  /* begin hardware dependent initializations */
  u32_t dwID;
  PCONP |= BIT_30; /* power up the mac */
  PINSEL2 |= (BIT_00 | BIT_02 | BIT_08 | BIT_16 | BIT_18 | BIT_20 |
    BIT_28 | BIT_30); /* select the rmii pins */
  PINSEL3 |= (BIT_00 | BIT_02); /* select the mdio and clk pins */
  MAC1 =  (BIT_08 | BIT_09 | BIT_10 | BIT_11 | BIT_14 | BIT_15); /* reset mac */
  Command = (BIT_03 | BIT_04 | BIT_05); /* reset the mac */
  MAC1 = BIT_01; /* pass all frames */
  MAC2 = (BIT_04 | BIT_05 | BIT_07); /* enable crc and padding */
  MAXF = MAX_FRAME; /* maximum frame length */
  CLRT = CLRT_DEFAULT; /* set the collision window and retransmission count */
  IPGR = IPGR_DEFAULT; /* set the back-to-back inter packet gap */
  MCFG = (BIT_02 | BIT_03 | BIT_04 | BIT_15); /* divide host clk by 20 and reset mii */
  MCFG ^= BIT_15; /* release the reset from the mii hardware */
  Command = (BIT_09 | BIT_06 /*| BIT_07*/); /* enable the rmii */
  SUPP = BIT_08; /* enable 100MBit mode */
  ethernetif_write_phy(PHY_BCR, BIT_15); /* reset the phy */
  while(ethernetif_read_phy(PHY_BCR) & (BIT_15 | BIT_11)); /* wait until reset done */
  ethernetif_write_phy(PHY_CTL, BIT_14); /* enable the link and act LEDs */
  dwID = (ethernetif_read_phy(PHY_ID1) << 16); /* get PHY manufacturer and revision */
  dwID |= (ethernetif_read_phy(PHY_ID2) & 0xFFF0);
	if(dwID != KSZ8001_ID) /* check if this is a KSZ8001L PHY */
    return enFalse; /* abort */
  ethernetif_write_phy(PHY_BCR, BIT_12); /* enable auto-negotiation */
  ethernetif_write_phy(PHY_ICS, BIT_08 | BIT_10); /* enable link up / down interrupt */
  PINSEL4 |= BIT_22; /* enable EINT1 interrupt */
  VICVectPriority15 = 0; /* set up the phy interrupt */
  VICVectAddr15 = (u32_t)ethernetif_linkinterrupt;
  VICIntEnable |= BIT_15; /* enable the link up/down interrupt */
  SA0 = (MAC_ADDRESS1 << 8) | MAC_ADDRESS2; /* set up the mac address */
  SA1 = (MAC_ADDRESS3 << 8) | MAC_ADDRESS4;
  SA2 = (MAC_ADDRESS5 << 8) | MAC_ADDRESS6;
  RxFilterCtrl = (BIT_01 | /*BIT_02 |*/ BIT_05); /* uni-, multi- and broadcast */
  init_rx_descriptors(); /* initialize rx descriptors */
  init_tx_descriptors(); /* initialize tx descriptors */
  Command |= (BIT_00 | BIT_01); /* enable receiver and transmitter */
  MAC1 |= BIT_00; /* enable the receiver */
  VICVectPriority21 = 0; /* set up the mac interrupt */
  VICVectAddr21 = (u32_t)ethernetif_interrupt;
  VICIntEnable |= BIT_21; /* enable the mac interrupt */
  /* end hardware dependent initializations */

  return ERR_OK;
}

void
ethernetif_poll(struct netif *netif)
{
  struct pbuf* p;
  p = low_level_input();
  if(p == NULL) return;
  netif->input(p, netif);
}

err_t
ethernetif_low_level_output(struct netif *netif, struct pbuf *p)
{
  struct pbuf* q;
	u32_t len;
  u32_t cpylen;
	u8_t error;

  u32_t txc_idx = TxConsumeIndex;
  u32_t txp_idx = TxProduceIndex;

  u8_t *rd_ptr; /* read pointer */
  u8_t *wr_ptr = (u8_t*)tx_descriptors[txp_idx * 2]; /* write pointer */

#if ETH_PAD_SIZE
  pbuf_header(p, -ETH_PAD_SIZE);			/* drop the padding word */
#endif

	len = 0;
	error = 0;
  for(q = p; q != NULL; q = q->next)
	{
		if((len + q->len) >= MAX_FRAME)
		{
			error = 1;
			break;
		}
    cpylen = q->len;
    rd_ptr = (u8_t*)q->payload;
    while(cpylen)
    {
      *wr_ptr = *rd_ptr;
      wr_ptr++; rd_ptr++;
      cpylen--;
    }
		len = len + q->len;
  }
	if(!error)
	{
    tx_descriptors[txp_idx * 2 + 1] = ((BIT_30 | BIT_29 | BIT_28 | BIT_26) | (len - 1));
    txp_idx++;
    if(txp_idx == MAC_TX_FRAGMENTS)
      txp_idx = 0;
    TxProduceIndex = txp_idx;
  }

#if ETH_PAD_SIZE
  pbuf_header(p, ETH_PAD_SIZE);
#endif

#if LINK_STATS
	if (!error)
		lwip_stats.link.xmit++;
#endif /* LINK_STATS */      

  return ERR_OK;
}

static void
ethernetif_linkinterrupt(void)
{
  volatile u16_t wInterrupt = ethernetif_read_phy(PHY_ICS); /* get the interrupt source */
  if(wInterrupt & BIT_00)
  {
    //if(ethernetif_read_phy(PHY_STATUS) & BIT_02) /* wait until link is up */
    //{
      if(ethernetif_read_phy(PHY_100BTX) & BIT_04) /* if full duplex is enabled */
      {
        MAC2 |= BIT_00; /* put the mac into full-duplex mode */
        Command |= BIT_10; /* issue full-duplex command */
        IPGT = 0x15; /* recommended setting; see LPC2468 User's Manual p. 223 */
      }
      else
      {
        IPGT = 0x12; /* recommended setting; see LPC2468 User's Manual p. 223 */
      }
      *status = status_connecting;
    //}
  }
  else if(wInterrupt & BIT_02)
  {
    /* do whatever is necessary when disconnected from the network */
    *status = status_disconnecting;
  }
  wInterrupt++;
  wInterrupt++;
  EXTINT |= BIT_01;
}

static void
ethernetif_interrupt(void)
{
}

static void
init_tx_descriptors(void)
{
  u16_t wIdx;
  u32_t* pdwTxDesc = tx_descriptors; /* get a pointer to the first element */
  for(wIdx = 0; wIdx < MAC_TX_FRAGMENTS; wIdx++)
  {
    *pdwTxDesc = (tDWord)&tx_buffer[wIdx * ETH_FRAG_SIZE]; /* packet */
    pdwTxDesc++;
    *pdwTxDesc = ((BIT_30 | BIT_29 | BIT_28 | BIT_26) | (ETH_FRAG_SIZE - 1)); /* control */
    pdwTxDesc++;
  }
  TxStatus = (u32_t)tx_status_descriptors;
  TxDescriptorNumber = MAC_TX_FRAGMENTS - 1; /* number of tx descriptors */
  TxProduceIndex = 0; /* begin at the first tx descriptor */
  TxDescriptor = (u32_t)tx_descriptors;
}

static void
init_rx_descriptors(void)
{
  u16_t wIdx;
  u32_t* pdwRxDesc = rx_descriptors; /* get a pointer to the firs element */
  for(wIdx = 0; wIdx < MAC_RX_FRAGMENTS; wIdx++)
  {
    *pdwRxDesc = (tDWord)&rx_buffer[wIdx * ETH_FRAG_SIZE]; /* packet */
    pdwRxDesc++;
    *pdwRxDesc = BIT_31 | (ETH_FRAG_SIZE - 1); /* control */
    pdwRxDesc++;
  }
  RxStatus = (u32_t)rx_status_descriptors;
  RxDescriptorNumber = MAC_RX_FRAGMENTS - 1; /* number of rx descriptors */
  RxConsumeIndex = 0; /* begin at the first rx descriptor */
  RxDescriptor = (u32_t)rx_descriptors;
}

static void
ethernetif_write_phy(tWord wReg, tWord wValue)
{
  MADR = (wReg & 0x1F); /* the register address. only bits 0..5 are valid. */
  MWTD = wValue; /* write data */
  while(MIND & BIT_00); /* wait until write cycle is complete */
}

static u16_t
ethernetif_read_phy(u16_t wReg)
{
  MADR = (wReg & 0x1F); /* the register address. only bits 0..5 are valid. */
  MCMD = BIT_00; /* issue a read command */
  while(MIND & BIT_00); /* wait until read operation is complete */
  MCMD = 0;
  return MRDD;
}

static struct pbuf*
low_level_input(void)
{
  struct pbuf *p, *q;
	u32_t len;
	u32_t cpylen;
  u32_t c;

  u8_t *rd_ptr; /* read pointer */
  u8_t *wr_ptr; /* write pointer */

  u32_t rxc_idx = RxConsumeIndex; /* just for the compiler... */
  u32_t rxp_idx = RxProduceIndex;

	if(rxc_idx == rxp_idx) /* check if a frame is available */
		return NULL;

	len = (rx_status_descriptors[rxc_idx * 2] & 0x7FF) - 3; /* get the packet size */
  rd_ptr = (u8_t*)rx_descriptors[rxc_idx * 2]; /* get the frame start address */

#if ETH_PAD_SIZE
  len += ETH_PAD_SIZE;						
#endif

  p = pbuf_alloc(PBUF_RAW, len, PBUF_POOL);
  if (p != NULL)
	{

#if ETH_PAD_SIZE
    pbuf_header(p, -ETH_PAD_SIZE);			
#endif

    for(q = p; q != NULL; q = q->next)
	  {
			if (q->len >= len)
				cpylen = len;
			else
				cpylen = q->len;
      c = cpylen;
      wr_ptr = (u8_t*)q->payload;
			while(c)
      {
        *wr_ptr = *rd_ptr;
        wr_ptr++; rd_ptr++;
        c--;
      }
			len = len - cpylen;
    }

    rxc_idx++;
    if(rxc_idx == MAC_RX_FRAGMENTS)
      rxc_idx = 0;
    RxConsumeIndex = rxc_idx;

#if ETH_PAD_SIZE
    pbuf_header(p, ETH_PAD_SIZE);			
#endif

#if LINK_STATS
    lwip_stats.link.recv++;
#endif

  }
	else 
	{
		rxc_idx++;
    if(rxc_idx == MAC_RX_FRAGMENTS)
      rxc_idx = 0;
    RxConsumeIndex = rxc_idx; /* drop the packet */

#if LINK_STATS
    lwip_stats.link.memerr++;
    lwip_stats.link.drop++;
#endif

  }
  return p;
}
